"""
This package implements Yandex.Dialogs request protocol.
The official documentation is available at
https://yandex.ru/dev/dialogs/alice/doc/protocol-docpage/#request
"""

import attr
from typing import Dict, List, Optional
from tgalice.utils.serialization import Serializeable, list_converter, dict_converter


class ENTITY_TYPES:
    YANDEX_GEO = 'YANDEX.GEO'
    YANDEX_FIO = 'YANDEX.FIO'
    YANDEX_NUMBER = 'YANDEX.NUMBER'
    YANDEX_DATETIME = 'YANDEX.DATETIME'


@attr.s
class Meta(Serializeable):
    locale: str = attr.ib()
    timezone: str = attr.ib()
    client_id: str = attr.ib()
    interfaces: dict = attr.ib(factory=dict)


@attr.s
class Span(Serializeable):
    start: int = attr.ib()
    end: int = attr.ib()


@attr.s
class Entity(Serializeable):
    type: str = attr.ib()  # may be one of ENTITY_TYPES, but not only
    tokens: Span = attr.ib(converter=Span.from_dict)
    value = attr.ib()


@attr.s
class Slot(Serializeable):
    type: str = attr.ib()
    tokens: Optional[Span] = attr.ib(converter=Span.from_dict, default=None)
    value = attr.ib(default=None)


@attr.s
class Intent(Serializeable):
    slots: Dict[str, Slot] = attr.ib(converter=dict_converter(Slot))


@attr.s
class NLU(Serializeable):
    tokens: List[str] = attr.ib(factory=list)
    entities: List[Entity] = attr.ib(converter=list_converter(Entity), factory=list)
    intents: Dict[str, Intent] = attr.ib(converter=dict_converter(Intent), factory=dict)


@attr.s
class Request(Serializeable):
    command: str = attr.ib(default=None)
    original_utterance: str = attr.ib(default=None)
    type: str = attr.ib(default=None)
    markup = attr.ib(factory=dict)
    payload = attr.ib(factory=dict)
    nlu: Optional[NLU] = attr.ib(converter=NLU.from_dict, default=None)


class User:
    user_id: str = attr.ib(default=None)
    access_token: str = attr.ib(default=None)


class Application:
    application_id: str = attr.ib(default=None)


@attr.s
class Session(Serializeable):
    message_id: int = attr.ib()
    session_id: str = attr.ib()
    skill_id: str = attr.ib()
    # user_id is deprecated, use application.application_id or user.user_id instead
    user_id: str = attr.ib()
    user: Optional[User] = attr.ib(default=None)
    application: Optional[Application] = attr.ib(default=None)
    new: bool = attr.ib(default=False)


@attr.s
class State(Serializeable):
    session: Optional[Dict] = attr.ib(default=None)
    user: Optional[Dict] = attr.ib(default=None)


@attr.s
class YandexRequest(Serializeable):
    meta: Meta = attr.ib(converter=Meta.from_dict)
    request: Request = attr.ib(converter=Request.from_dict)
    session: Optional[Session] = attr.ib(converter=Session.from_dict)
    new: bool = attr.ib(default=False)
    version: str = attr.ib(default='1.0')
    state: Optional[State] = attr.ib(converter=State.from_dict, default=None)



"""
Automaton dialog manager is based on several concepts:
- phrase: a template, based on which the current response can be generated
- intent: a class with which user input (not only text!) can be matched
- state: a node that defines current phrase and allowed next transitions
- transition: pair (intent, next state, [prefix phrase, condition])

"""
import attr
import copy

from collections import Counter, defaultdict, OrderedDict
from collections.abc import Mapping
from typing import Dict, Optional, Tuple

from tgalice.dialog import Context, Response
from tgalice.dialog.phrase import Phrase
from tgalice.nlu.matchers import make_matcher, RegexMatcher
from tgalice.utils.configuration import load_config

from tgalice.dialog_manager.base import CascadableDialogManager


ANY_STATE = 'universal_state'


class State:
    def __init__(self, name, phrase, additional_suggests=None, restore_prev_state=False):
        self.name = name
        self.phrase = phrase
        self.additional_suggests = additional_suggests or []
        self.restore_prev_state = restore_prev_state  # useful for intents like contextual help
        # todo: choose the phrase depending on the prev state (or some other condition)


class Intent:
    def __init__(self, name, regex=None, examples=None, threshold=None):
        self.name = name
        self.regex = regex
        self.examples = examples or []
        self.threshold = threshold  # custom threshold for matcher

    def __repr__(self):
        return 'Intent({})'.format(self.name)


class Transition:
    def __init__(self, name, intent, prev_state, next_state, priority=1):
        # todo: add different transition conditions, beside intents: e.g. callback_data or custom_filter
        self.name = name
        self.intent = intent
        self.prev_state = prev_state
        self.next_state = next_state
        self.priority = priority

    def __repr__(self):
        return 'Transition({}, {}, {}, {})'.format(self.name, self.intent, self.prev_state, self.next_state)


@attr.s
class FSAConfigOptions:
    name: str = attr.ib(default='default_automaton')
    initial_state: str = attr.ib(default=None)
    state_on_new_session: str = attr.ib(default=None)

    @classmethod
    def from_dict(cls, data=None):
        if isinstance(data, cls):
            return data
        return cls(**(data or {}))


@attr.s
class FSAConfig:
    states = attr.ib(factory=dict)
    intents = attr.ib(factory=dict)
    transitions = attr.ib(factory=dict)
    phrases = attr.ib(factory=dict)
    options = attr.ib(converter=FSAConfigOptions.from_dict, factory=FSAConfigOptions.from_dict)


class AutomatonDialogManager(CascadableDialogManager):
    def __init__(self, config, matcher='exact', match_score_first=False, **kwargs):
        super(AutomatonDialogManager, self).__init__(**kwargs)
        if isinstance(matcher, str):
            matcher = make_matcher(matcher)
        elif isinstance(matcher, Mapping):
            matcher = make_matcher(**matcher)
        self.matcher = matcher
        self.regex_matcher = RegexMatcher()
        self.match_score_first = match_score_first

        if not isinstance(config, FSAConfig):
            config = FSAConfig(**load_config(config))
        self._cfg = config
        if not self._cfg.states:
            raise ValueError('The "states" config property cannot be empty.')

        self.phrases: OrderedDict[str, Phrase] = OrderedDict()
        self.intents: OrderedDict[str, Intent] = OrderedDict()
        self.states: OrderedDict[str, State] = OrderedDict()
        self.transitions: OrderedDict[str, Transition] = OrderedDict()
        self.intent2transition = defaultdict(list)

        self.universal_state = self.add_state(name=ANY_STATE, phrase=None)
        # todo: make sure this state is never the destination

        for phrase_name, kw in self._cfg.phrases.items():
            self.add_phrase(name=phrase_name, **kw)
        for intent_name, kw in self._cfg.intents.items():
            self.add_intent(name=intent_name, **kw)
        for state_name, kw in self._cfg.states.items():
            self.add_state(name=state_name, **kw)
        for transition_name, kw in self._cfg.transitions.items():
            self.add_transition(name=transition_name, **kw)

        # the first state is Universal and should be skipped
        self.initial_state = self._cfg.options.initial_state or list(self.states.keys())[1]
        self.state_on_new_session = self._cfg.options.state_on_new_session or self.initial_state
        self.name = self._cfg.options.name

        self.validate()

        self.compile_matchers()

    @property
    def _collections(self):
        return {
            'intent': self.intents,
            'phrase': self.phrases,
            'transition': self.transitions,
            'state': self.states,
        }

    def _check_name(self, entity_type, name, kwargs):
        collection = self._collections[entity_type]
        if name is None:
            if 'label' in kwargs:
                name = kwargs['label']
                del kwargs['label']
            else:
                name = '{}__{}'.format(entity_type, len(collection))
        assert name not in collection
        return name

    def add_state(self, name=None, **kwargs) -> State:
        kwargs = copy.deepcopy(kwargs)
        name = self._check_name('state', name, kwargs)
        if 'q' in kwargs:
            intent = self.add_intent(name=name, examples=kwargs['q'])
            del kwargs['q']
            self.add_transition(
                intent=intent.name,
                prev_state=ANY_STATE,
                next_state=name,
            )
        if 'a' in kwargs:
            phrase = self.add_phrase(name=name, text=kwargs['a'])
            del kwargs['a']
            kwargs['phrase'] = phrase.name
        if 'next' in kwargs:
            for transition_args in kwargs['next']:
                if 'label' in transition_args:
                    if 'intent' in transition_args:
                        new_intent_name = transition_args['intent']
                    else:
                        examples = transition_args.get('examples', [])
                        if 'suggest' in transition_args:
                            examples.append(transition_args['suggest'])
                        new_intent = self.add_intent(
                            regex=transition_args.get('regex'),
                            examples=examples,
                        )
                        new_intent_name = new_intent.name
                    self.add_transition(
                        intent=new_intent_name,
                        prev_state=name,
                        next_state=transition_args['label'],
                    )
                if 'suggest' in transition_args:
                    if 'additional_suggests' not in kwargs:
                        kwargs['additional_suggests'] = []
                    kwargs['additional_suggests'].append(transition_args['suggest'])
            del kwargs['next']
        if 'default_next' in kwargs:
            self.add_transition(
                intent=None,  # todo: add special tags to transitions-without-intents
                prev_state=name,
                next_state=kwargs['default_next'],
            )
            del kwargs['default_next']

        state = State(name=name, **kwargs)
        self.states[name] = state
        return state

    def add_phrase(self, name=None, **kwargs) -> Phrase:
        kwargs = copy.deepcopy(kwargs)
        name = self._check_name('phrase', name, kwargs)
        phrase = Phrase(name=name, **kwargs)
        self.phrases[name] = phrase
        return phrase

    def add_intent(self, name=None, **kwargs) -> Intent:
        kwargs = copy.deepcopy(kwargs)
        name = self._check_name('intent', name, kwargs)
        intent = Intent(name=name, **kwargs)
        self.intents[name] = intent
        return intent

    def add_transition(self, name=None, **kwargs) -> Transition:
        kwargs = copy.deepcopy(kwargs)
        name = self._check_name('transition', name, kwargs)
        transition = Transition(name=name, **kwargs)
        self.transitions[name] = transition
        self.intent2transition[transition.intent].append(transition.name)
        return transition

    def validate(self, silent=False):
        errors = []
        for transition_name, transition in self.transitions.items():
            if transition.prev_state not in self.states:
                errors.append('State {} not found'.format(transition.prev_state))
            if transition.next_state not in self.states:
                errors.append('State {} not found'.format(transition.next_state))
            if transition.intent not in self.intents and transition.intent is not None:
                errors.append('Intent {} not found'.format(transition.intent))
        if errors and not silent:
            raise ValueError('\n'.join(['Some transitions were invalid. Errors:'] + errors))
        return errors

    def compile_matchers(self):
        texts = []
        labels = []
        for intent_name, intent in self.intents.items():
            for example in intent.examples:
                texts.append(example)
                labels.append(intent.name)
            if intent.threshold:
                self.matcher.thresholds[intent_name] = intent.threshold
        self.matcher.fit(texts, labels)

        texts = []
        labels = []
        for intent_name, intent in self.intents.items():
            if not intent.regex:
                continue
            regex = intent.regex
            if not isinstance(regex, list):
                regex = [regex]
            for example in regex:
                texts.append(example)
                labels.append(intent.name)
        self.regex_matcher.fit(texts, labels)

    def extract_prev_state(self, context: Context) -> Optional[State]:
        state_name = context.user_object.get('automaton', {}).get(self.name, {}).get('state_name')
        if state_name:
            return self.states[state_name]

    def try_to_respond(self, ctx: Context) -> Optional[Response]:
        prev_state = self.extract_prev_state(ctx)
        if prev_state is None:
            new_state_name, updated_user_object = self.initialize(ctx)
        else:
            new_state_name, updated_user_object = self.do_transition(prev_state, ctx)
        if new_state_name is None:
            # todo: maybe do some default action
            pass
        if new_state_name is not None:
            new_state = self.states[new_state_name]
            phrase = self.phrases[new_state.phrase]
            response = phrase.render(additional_suggests=new_state.additional_suggests)
            if prev_state and new_state.restore_prev_state:
                self.remember_new_state(prev_state.name, updated_user_object)
            response.updated_user_object = updated_user_object
            return response
        return None

    def initialize(self, context) -> Tuple[str, dict]:
        user_object = context.user_object
        self.remember_new_state(self.initial_state, user_object)
        return self.initial_state, user_object

    def get_intent_scores(self, context: Context) -> Counter:
        scores = self.matcher.aggregate_scores(context.message_text)
        scores.update(self.regex_matcher.aggregate_scores(context.message_text))
        if context.yandex and context.yandex.request.nlu:
            for intent in context.yandex.request.nlu.intents:
                scores[intent] = 1
        return scores

    def do_transition(self, prev_state: State, context: Context) -> Tuple[Optional[str], Optional[dict]]:
        if context.message_text:
            scores = self.get_intent_scores(context=context)
            next_state_name, match_score = None, None
            # todo: реализовать сортировку в любом порядке: скор, номер перехода, приоритет
            for intent_name, score in scores.most_common():
                next_state_name = self.find_transition_by_intent(intent_name, prev_state)
            if next_state_name is None:
                # the transitions-without-text have lower priority than any text (not sure)
                next_state_name = self.find_transition_by_intent(None, prev_state)
            if next_state_name is None:
                return None, None
            user_object = context.user_object
            self.remember_new_state(next_state_name, user_object)
            return next_state_name, user_object
        elif context.session_is_new():
            next_state_name = self.state_on_new_session
            user_object = context.user_object
            self.remember_new_state(next_state_name, user_object)
            return next_state_name, user_object
        else:
            # todo: find a convenient way to match non-text inputs
            return None, None

    def remember_new_state(self, new_state_name: str, user_object: Optional[Mapping]):
        if user_object is None:
            user_object = {}
        if 'automaton' not in user_object:
            user_object['automaton'] = {}
        if self.name not in user_object['automaton']:
            user_object['automaton'][self.name] = {}
        user_object['automaton'][self.name]['state_name'] = new_state_name
        return user_object

    def find_transition_by_intent(self, intent, prev_state: State):
        for transition_name in self.intent2transition[intent]:
            transition = self.transitions[transition_name]
            if transition.prev_state != prev_state.name and transition.prev_state != ANY_STATE:
                continue
            return transition.next_state



import attr
import requests


from typing import Dict, List, Optional


@attr.s
class Image:
    id: str = attr.ib()
    size: int = attr.ib()
    createdAt: str = attr.ib()
    origUrl: Optional[str] = attr.ib(default=None)


class YandexImageAPI:
    """
    This class is a wrapper around Yandex image storage API.
    Its official documentation is available online:
    https://yandex.ru/dev/dialogs/alice/doc/resource-upload-docpage/
    """
    def __init__(self, token, skill_id, default_image_id=None, upload_just_in_time=False):
        self.token = token
        self.skill_id = skill_id
        self.default_image_id = default_image_id
        self.upload_just_in_time = upload_just_in_time
        self.url2image: Dict[str, Image] = {}
        self.id2image: Dict[str, Image] = {}

    def update_images(self) -> None:
        """ Retrieve the list of images from the cloud storage and save it to the local index. """
        for image in self.get_images_list():
            if image.origUrl:
                self.url2image[image.origUrl] = image
            self.id2image[image.id] = image

    def add_image(self, url, timeout=5) -> Optional[Image]:
        """ Add image to the local index and Yandex storage by its url."""
        if url in self.url2image:
            return self.url2image[url]
        result = self.upload_image(url, timeout=timeout)
        if result:
            self.url2image[url] = result
            self.id2image[result.id] = result
        return result

    def upload_image(self, url, timeout=5) -> Optional[Image]:
        """
        Try to upload the image by url (without adding it to the local index)
        small images take 1.5-2 seconds to upload
        """
        r = requests.post(
            url='https://dialogs.yandex.net/api/v1/skills/{}/images'.format(self.skill_id),
            headers={'Authorization': 'OAuth {}'.format(self.token)},
            json={'url': url},
            timeout=timeout,
        )
        result = r.json().get('image')
        if result:
            return Image(**result)

    def get_images_list(self) -> List[Image]:
        """ Get all images in the Yandex storage. """
        r = requests.get(
            url='https://dialogs.yandex.net/api/v1/skills/{}/images'.format(self.skill_id),
            headers={'Authorization': 'OAuth {}'.format(self.token)}
        )
        results = r.json().get('images', [])
        return [Image(**item) for item in results]

    def get_image_id_by_url(self, url, try_upload=None, timeout=2) -> Optional[str]:
        """
        Try to get image id from local storage or quickly upload it
        or return the default image.
        """
        if url in self.url2image:
            return self.url2image[url].id
        if try_upload is None:
            try_upload = self.upload_just_in_time
        if try_upload:
            image = self.add_image(url, timeout=timeout)
            if image:
                return image.id
        if self.default_image_id:
            return self.default_image_id

    def get_quota(self):
        """ Get existing an occupied amount of storage for images and sounds in bytes"""
        r = requests.get(
            url='https://dialogs.yandex.net/api/v1/status',
            headers={'Authorization': 'OAuth {}'.format(self.token)}
        )
        return r.json()

    def delete_image(self, image_id):
        """ Delete image from storage by its id and delete it from local index """
        r = requests.delete(
            url='https://dialogs.yandex.net/api/v1/skills/{}/images/{}'.format(self.skill_id, image_id),
            headers={'Authorization': 'OAuth {}'.format(self.token)}
        )
        if r.ok:
            if image_id in self.id2image:
                image = self.id2image[image_id]
                del self.id2image[image_id]
                if image.origUrl in self.url2image and self.url2image[image.origUrl].id == image_id:
                    del self.url2image[image.origUrl]
        return r.json()



# coding: utf-8
import typing

from ..nlu import basic_nlu
from ..dialog import Context, Response
from ..dialog.names import COMMANDS


class BaseDialogManager:
    """ This class defines the interface of dialog managers with a single `response` function. """
    def __init__(self, default_message='I dont\'t understand.'):
        self.default_message = default_message

    def respond(self, ctx: Context):
        return Response(text=self.default_message)


class CascadableDialogManager(BaseDialogManager):
    """ This interface allows dialog manager to have no answer - and to let other managers try.
    The `method try_to_respond` is like `respond`, but may also return None.
    It is expected to be used with CascadeDialogManager.
    """
    def __init__(self, *args, **kwargs):
        super(CascadableDialogManager, self).__init__(*args, **kwargs)

    def try_to_respond(self, ctx: Context) -> typing.Union[Response, None]:
        """ This method should return None or a valid Response """
        raise NotImplementedError()

    def respond(self, ctx):
        response = self.try_to_respond(ctx)
        if isinstance(response, Response):
            return response
        return Response(text=self.default_message)


class CascadeDialogManager(BaseDialogManager):
    """ This dialog manager tries multiple dialog managers in turn, and returns the first successful response. """
    def __init__(self, *managers):
        super(CascadeDialogManager, self).__init__()
        assert len(managers) > 0
        for manager in managers[:-1]:
            assert isinstance(manager, CascadableDialogManager)
        self.candidates = managers[:-1]
        assert isinstance(managers[-1], BaseDialogManager)
        self.final_candidate = managers[-1]

    def respond(self, *args, **kwargs):
        for manager in self.candidates:
            response = manager.try_to_respond(*args, **kwargs)
            if response is not None:
                return response
        return self.final_candidate.respond(*args, **kwargs)


class GreetAndHelpDialogManager(CascadableDialogManager):
    """ This dialog manager can be responsible for the first and the last messages, and for the help message. """
    def __init__(self, greeting_message, help_message, *args, exit_message=None, **kwargs):
        super(GreetAndHelpDialogManager, self).__init__(*args, **kwargs)
        self.greeting_message = greeting_message
        self.help_message = help_message
        self.exit_message = exit_message

    def try_to_respond(self, ctx: Context):
        if self.is_first_message(ctx):
            return Response(text=self.greeting_message)
        if self.is_like_help(ctx):
            return Response(text=self.help_message)
        if self.exit_message is not None and self.is_like_exit(ctx):
            return Response(text=self.exit_message, commands=[COMMANDS.EXIT])
        return None

    def is_first_message(self, context):
        if not context.message_text or context.message_text == '/start':
            return True
        if basic_nlu.like_help(context.message_text):
            return True
        return False

    def is_like_help(self, context):
        return basic_nlu.like_help(context.message_text)

    def is_like_exit(self, context):
        return basic_nlu.like_exit(context.message_text)



from tgalice.nlg.reply_markup import TTSParser
from tgalice.dialog_manager.base import Response

PARSED_RESPONSE = dict(
    markup='This markup be used in <text>Yandex.Dialogs</text><voice>Yandex Dialogs</voice> and other platforms.'
           '<a href="https://dialogs.yandex.ru/">Yandex.Dialogs site</a>',
    expected_text='This markup be used in Yandex.Dialogs and other platforms.',
    expected_voice='This markup be used in Yandex Dialogs and other platforms.',
    expected_links=[{'title': 'Yandex.Dialogs site', 'url': 'https://dialogs.yandex.ru/'}],
)


def test_tts_parser():
    parser = TTSParser()
    parser.feed(PARSED_RESPONSE['markup'])
    assert parser.get_text() == PARSED_RESPONSE['expected_text']
    assert parser.get_voice() == PARSED_RESPONSE['expected_voice']
    assert parser.get_links() == PARSED_RESPONSE['expected_links']


def test_response_set_text():
    resp = Response(text=None).set_text(PARSED_RESPONSE['markup'])
    assert resp.text == PARSED_RESPONSE['expected_text']
    assert resp.voice == PARSED_RESPONSE['expected_voice']
    assert resp.links == PARSED_RESPONSE['expected_links']



import copy
import logging

from datetime import datetime

from tgalice.dialog import Context, Response
from tgalice.dialog.names import SOURCES
from tgalice.storage.database_utils import get_mongo_or_mock, replace_dotted_keys


try:
    import pymongo
except ModuleNotFoundError:
    pymongo = None


logger = logging.getLogger(__name__)


class LoggedMessage:
    def __init__(self, text, user_id, from_user, **kwargs):
        self.text = text
        self.user_id = user_id
        self.from_user = from_user
        self.timestamp = str(datetime.utcnow())
        self.kwargs = kwargs
        """
        Expected kwargs:
            text
            user_id
            message_id
            from_user
            username
            reply_to_id
            source
            data        (original message in Alice)
            label       (something like intent)
            request_id  (this id the same for request and response, useful for joining logs)
        """

    def save_to_mongo(self, collection):
        collection.insert_one(replace_dotted_keys(self.to_dict()))

    def to_dict(self):
        result = {
            'text': self.text,
            'user_id': self.user_id,
            'from_user': self.from_user,
            'timestamp': self.timestamp
        }
        for k, v in self.kwargs.items():
            if k not in result:
                result[k] = v
        return result

    @classmethod
    def from_context(cls, context: Context, **kwargs):
        serializable_message = context.raw_message
        if context.source == SOURCES.TELEGRAM:
            message = context.raw_message
            if message.reply_to_message is not None:
                kwargs['reply_to_id'] = message.reply_to_message.message_id
            kwargs['message_id'] = message.message_id
            kwargs['username'] = message.chat.username
            serializable_message = {'message': str(message)}
        elif context.source == SOURCES.VK:
            serializable_message = {'message': context.raw_message.to_json()}
        if context.request_id is not None:
            kwargs['request_id'] = context.request_id
        return cls(
            text=context.message_text,
            user_id=context.user_id,
            from_user=True,
            data=serializable_message,
            source=context.source,
            **kwargs
        )

    @classmethod
    def from_response(cls, data, context: Context, response: Response, **kwargs):
        data = copy.deepcopy(data)
        if context.source == SOURCES.TELEGRAM:
            message = context.raw_message
            kwargs['reply_to_id'] = message.message_id
            kwargs['username'] = message.chat.username
            # todo: maybe somehow get message_id for output messages
            if 'reply_markup' in data:
                data['reply_markup'] = data['reply_markup'].to_json()
        if context.request_id is not None:
            kwargs['request_id'] = context.request_id
        return cls(
            text=response.text,
            user_id=context.user_id,
            from_user=False,
            data=data,
            source=context.source,
            **kwargs
        )


class BaseMessageLogger:
    def __init__(self, detect_pings=False, not_log_id=None):
        self.detect_pings = detect_pings
        self.not_log_id = not_log_id or set()

    def log_context(self, context, **kwargs):
        return self.log_message(message=context.raw_message, context=context, source=context.source, **kwargs)

    def log_response(self, data, source, context=None, response=None, **kwargs):
        return self.log_message(message=data, source=source, context=context, response=response, **kwargs)

    def log_message(self, message, source, context=None, response=None, **kwargs):
        if context is not None:
            if response is not None:
                msg = LoggedMessage.from_response(message, context=context, response=response)
            else:
                msg = LoggedMessage.from_context(context)
        else:
            return
        if response is not None and response.label is not None:
            msg.kwargs['label'] = response.label
        if self.not_log_id is not None and msg.user_id in self.not_log_id:
            # main reason: don't log pings from Yandex
            return
        if self.detect_pings and self.is_like_ping(context):
            return
        self.save_a_message(msg.to_dict())

    def is_like_ping(self, context=None):
        return context is not None and context.source == SOURCES.ALICE \
               and context.message_text == 'ping' and context.session_is_new()

    def save_a_message(self, message_dict):
        logger.info(message_dict)


class MongoMessageLogger(BaseMessageLogger):
    def __init__(self, collection=None, database=None, collection_name='message_logs', write_concern=0, **kwargs):
        super(MongoMessageLogger, self).__init__(**kwargs)
        self.collection = collection
        if self.collection is None:
            if database is None:
                database = get_mongo_or_mock()
            if pymongo and not isinstance(write_concern, pymongo.write_concern.WriteConcern):
                write_concern = pymongo.write_concern.WriteConcern(w=write_concern)
            self.collection = database.get_collection(collection_name, write_concern=write_concern)

    def save_a_message(self, message_dict):
        self.collection.insert_one(replace_dotted_keys(message_dict))




class SOURCES:
    ALICE = 'alice'
    FACEBOOK = 'facebook'
    TELEGRAM = 'telegram'
    TEXT = 'text'
    VK = 'vk'
    unknown_source_error_message = 'Source must be on of {"alice", "facebook", "telegram", "text", "vk"}'


class COMMANDS:
    EXIT = 'exit'


class REQUEST_TYPES:
    SIMPLE_UTTERANCE = 'SimpleUtterance'
    BUTTON_PRESSED = 'ButtonPressed'



from __future__ import print_function

import argparse
import json
import logging
import os
import telebot
import warnings

from flask import Flask, request
from pymessenger.bot import Bot as FacebookBot

from tgalice.interfaces.vk import VKBot, VKMessage
from tgalice.dialog_connector import DialogConnector, SOURCES


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class FlaskServer:
    def __init__(
            self,
            connector: DialogConnector,
            telegram_token=None,
            facebook_access_token=None,
            facebook_verify_token=None,
            base_url=None,
            alice_url='alice/', telegram_url='tg/', facebook_url='fb/', vk_url='vk/',
            restart_webhook_url='restart_webhook',
            vk_token=None,
            vk_group_id=None,
            app=None,
    ):
        self.telegram_token = telegram_token or os.environ.get('TOKEN') or os.environ.get('TELEGRAM_TOKEN')
        self.facebook_access_token = facebook_access_token or os.environ.get('FACEBOOK_ACCESS_TOKEN')
        self.facebook_verify_token = facebook_verify_token or os.environ.get('FACEBOOK_VERIFY_TOKEN')
        self.vk_token = vk_token or os.environ.get('VK_TOKEN')
        self.vk_group_id = vk_group_id or os.environ.get('VK_GROUP_ID')
        if base_url is None:
            base_url = os.environ.get('BASE_URL')
        self.base_url = base_url
        self.alice_url = alice_url
        self.telegram_url = telegram_url
        self.facebook_url = facebook_url
        self.vk_url = vk_url
        self.restart_webhook_url = restart_webhook_url

        self.connector = connector

        self.app = app or Flask(__name__)

        logger.info('The Alice webhook is available on "{}"'.format(self.alice_webhook_url))
        self.app.route(self.alice_webhook_url, methods=['POST'])(self.alice_response)

        if self.telegram_token is not None:
            self.bot = telebot.TeleBot(self.telegram_token)
            self.bot.message_handler(func=lambda message: True)(self.tg_response)
            if base_url is not None:
                logger.info('Running Telegram bot with token "{}" on "{}"'.format(
                    self.telegram_token, self.telegram_webhook_url)
                )
                self.app.route(self.telegram_webhook_url, methods=['POST'])(self.get_tg_message)
                self.app.route("/" + self.restart_webhook_url)(self.telegram_web_hook)
            else:
                logger.info(
                    'Running Telegram bot with token "{}", but without BASE_URL it can work only locally'.format(
                        self.telegram_token
                    )
                )
        else:
            logger.info('Running no Telegram bot because TOKEN or BASE_URL was not provided')
            self.bot = None

        if self.facebook_verify_token and self.facebook_access_token:
            logger.info('Running Facebook bot on "{}"'.format(self.facebook_webhook_url))
            self.app.route(self.facebook_webhook_url, methods=['GET'])(self.receive_fb_verification_request)
            self.app.route(self.facebook_webhook_url, methods=['POST'])(self.facebook_response)
            self.facebook_bot = FacebookBot(self.facebook_access_token)
        else:
            logger.info(
                'Running no Facebook bot because FACEBOOK_ACCESS_TOKEN or FACEBOOK_VERIFY_TOKEN was not provided'
            )
            self.facebook_bot = None

        self.vk_bot = None
        if self.vk_token is None:
            logger.info('Skipping VK setup because vk_token is empty')
        elif self.vk_group_id is None:
            logger.info('Skipping VK setup because vk_group_id is empty')
        else:
            self.vk_bot = VKBot(token=self.vk_token, group_id=self.vk_group_id)
            self.vk_bot.message_handler()(self.vk_response)
            self.app.route("/" + self.vk_url, methods=['POST'])(lambda: self.vk_bot.process_webhook_data(request.json))

        self._processed_telegram_ids = set()

    @property
    def alice_webhook_url(self):
        return "/" + self.alice_url

    @property
    def facebook_webhook_url(self):
        return '/' + self.facebook_url

    @property
    def telegram_webhook_url(self):
        return '/' + self.telegram_url + self.telegram_token

    def alice_response(self):
        logger.info('Got message from Alice: {}'.format(request.json))
        response = self.connector.respond(request.json, source=SOURCES.ALICE)
        logger.info('Sending message to Alice: {}'.format(response))
        return json.dumps(response, ensure_ascii=False, indent=2)

    def tg_response(self, message):
        logger.info('Got message from Telegram: {}'.format(message))
        if message.message_id in self._processed_telegram_ids:
            # avoid duplicate response after the bot starts
            logger.info('Telegram message id {} is duplicate, skipping it'.format(message.message_id))
            return
        self._processed_telegram_ids.add(message.message_id)
        # todo: cleanup old ids from _processed_telegram_ids
        response = self.connector.respond(message, source=SOURCES.TELEGRAM)
        telegram_response = self.bot.reply_to(message, **response)
        multimedia = response.pop('multimedia', [])
        for item in multimedia:
            if item['type'] == 'photo':
                self.bot.send_photo(message.chat.id, photo=item['content'], **response)
            if item['type'] == 'document':
                self.bot.send_document(message.chat.id, data=item['content'], **response)
            elif item['type'] == 'audio':
                self.bot.send_audio(message.chat.id, audio=item['content'], **response)
        logger.info('Sent a response to Telegram: {}'.format(message))

    def vk_response(self, message: VKMessage):
        response = self.connector.respond(message, source=SOURCES.VK)
        self.vk_bot.send_message(user_id=message.user_id, **response)

    def get_tg_message(self):
        self.bot.process_new_updates([telebot.types.Update.de_json(request.stream.read().decode("utf-8"))])
        return "!", 200

    def telegram_web_hook(self):
        self.bot.remove_webhook()
        self.bot.set_webhook(url=self.base_url + self.telegram_url + self.telegram_token)
        return "Telegram webhook restarted!", 200

    def vk_web_hook(self):
        endpoint = self.base_url + '/' + self.vk_url
        logger.info('Setting vk webhook on {}'.format(endpoint))
        self.vk_bot.set_postponed_webhook(url=endpoint, remove_old=True)

    def run_local_telegram(self):
        if self.bot is not None:
            self.bot.remove_webhook()
            self.bot.polling()
        else:
            raise ValueError('Cannot run Telegram bot, because Telegram token was not found.')

    def run_local_vk(self):
        if self.vk_bot is not None:
            self.vk_bot.remove_webhook()
            self.vk_bot.polling()
        else:
            raise ValueError('VK bot is not initialized and cannot run.')

    def receive_fb_verification_request(self):
        """Before allowing people to message your bot, Facebook has implemented a verify token
        that confirms all requests that your bot receives came from Facebook."""
        token_sent = request.args.get("hub.verify_token")
        if token_sent == self.facebook_verify_token:
            return request.args.get("hub.challenge")
        return 'Invalid verification token'

    def facebook_response(self):
        output = request.get_json()
        logger.info('Got messages from Facebook: {}'.format(output))
        for event in output['entry']:
            messaging = event['messaging']
            for message in messaging:
                if message.get('message') or message.get('postback'):
                    recipient_id = message['sender']['id']
                    response = self.connector.respond(message, source=SOURCES.FACEBOOK)
                    logger.info('Sending message to Facebook: {}'.format(response))
                    self.facebook_bot.send_message(recipient_id, response)
        return "Message Processed"

    def run_server(self, host="0.0.0.0", port=None, use_ngrok=False):
        # todo: maybe, run a foreign app instead (attach own blueprint to it)
        if port is None:
            port = int(os.environ.get('PORT', 5000))
        if use_ngrok:
            from tgalice.server.flask_ngrok import run_ngrok
            logger.info('starting ngrok...')
            ngrok_address = run_ngrok(port)
            logger.info('ngrok_address is {}'.format(ngrok_address))
            self.base_url = ngrok_address
        if self.telegram_token is not None and self.base_url is not None:
            self.telegram_web_hook()
        else:
            warnings.warn('Either telegram token or base_url was not found; cannot run Telegram bot.')
        if self.vk_bot:
            self.vk_web_hook()
        else:
            logging.warning('VK bot has not been created; cannot run it')
        self.app.run(host=host, port=port)

    def run_command_line(self):
        input_sentence = ''
        while True:
            response, need_to_exit = self.connector.respond(input_sentence, source=SOURCES.TEXT)
            print(response)
            if need_to_exit:
                break
            input_sentence = input('> ')

    def parse_args_and_run(self):
        parser = argparse.ArgumentParser(description='Run the bot')
        parser.add_argument('--cli', action='store_true', help='Run the bot locally in command line mode')
        parser.add_argument('--poll', action='store_true', help='Run the bot locally in polling mode (Telegram or VK)')
        parser.add_argument('--ngrok', action='store_true',
                            help='Run the bot locally with ngrok tunnel into the Internet')
        args = parser.parse_args()
        if args.cli:
            self.run_command_line()
        elif args.poll:
            if self.bot:
                self.run_local_telegram()
            elif self.vk_bot:
                self.run_local_vk()
            else:
                raise ValueError('Got no local bots to run in polling mode')
        else:
            self.run_server(use_ngrok=args.ngrok)



import copy
import logging
import uuid

from typing import Optional

from .names import REQUEST_TYPES, SOURCES
from tgalice.interfaces.yandex import YandexRequest


logger = logging.getLogger(__name__)


class Context:
    def __init__(
            self, user_object, message_text, metadata, request_id=None, user_id=None, source=None, raw_message=None,
            request_type=REQUEST_TYPES.SIMPLE_UTTERANCE, payload=None, yandex=None,
    ):
        self._user_object = copy.deepcopy(user_object)
        self.message_text = message_text
        self.metadata = metadata
        self.request_id = request_id or str(uuid.uuid1())
        self.user_id = user_id
        self.source = source
        self.raw_message = raw_message
        self.request_type = request_type
        self.payload = payload
        self.yandex: Optional[YandexRequest] = yandex

    @property
    def user_object(self):
        # todo: make _user_object constant instead of copying it every time
        return copy.deepcopy(self._user_object)

    def add_user_object(self, user_object):
        self._user_object = copy.deepcopy(user_object)

    def session_is_new(self):
        # todo: define new session for non-Alice sources as well
        return bool(self.metadata.get('new_session'))

    @classmethod
    def from_raw(cls, source, message):
        metadata = {}
        if source == SOURCES.TELEGRAM:
            user_id = source + '__' + str(message.from_user.id)
            message_text = message.text
        elif source == SOURCES.VK:
            user_id = source + '__' + str(message.user_id)
            message_text = message.text
        elif source == SOURCES.ALICE:
            sess = message['session']
            if sess.get('user', {}).get('user_id'):
                # the new user_id, which is persistent across applications
                user_id = source + '_auth__' + sess['user']['user_id']
            else:
                # the old user id, that changes across applications
                user_id = source + '__' + sess['user_id']
            message_text = message['request'].get('command', '')
            metadata['new_session'] = message.get('session', {}).get('new', False)
        elif source == SOURCES.FACEBOOK:
            user_id = source + '__' + message['sender']['id']
            message_text = message.get('message', {}).get('text', '')
        elif source == SOURCES.TEXT:
            user_id = 'the_text_user'
            message_text = message
        else:
            raise ValueError(SOURCES.unknown_source_error_message + ', got {} instead'.format(source))
        ctx = cls(
            user_object=None,
            message_text=message_text,
            metadata=metadata,
            user_id=user_id,
            source=source,
            raw_message=message,
        )
        if source == SOURCES.ALICE:
            ctx.request_type = message['request'].get('type', REQUEST_TYPES.SIMPLE_UTTERANCE)
            ctx.payload = message['request'].get('payload', {})
            try:
                ctx.yandex = YandexRequest.from_dict(message)
            except Exception as e:
                logger.warning('Could not deserialize Yandex request: got exception "{}".'.format(e))
        elif source == SOURCES.FACEBOOK:
            if not message.get('message', {}).get('text', ''):
                payload = message.get('postback', {}).get('payload')
                if payload is not None:
                    ctx.payload = payload
                    ctx.request_type = REQUEST_TYPES.BUTTON_PRESSED  # todo: check if it is really the case
                # todo: do something in case of attachments (message['message'].get('attachments'))
                # if user sends us a GIF, photo,video, or any other non-text item
        return ctx



import numpy as np
import pytest
import math


from tgalice.nlu import matchers

sample_texts = ['привет', 'добрый день', 'сколько времени']
sample_labels = ['hello', 'hello', 'get_time']

NO_MATCH = (None, -math.inf)


def test_exact_matcher():
    matcher = matchers.make_matcher('exact')
    matcher.fit(sample_texts, sample_labels)
    assert matcher.match('привет') == ('hello', 1)
    assert matcher.match('приветик') == NO_MATCH
    assert matcher.match('добрый день') == ('hello', 1)
    assert matcher.match('день добрый') == NO_MATCH


def test_jaccard_matcher():
    matcher = matchers.JaccardMatcher(threshold=0.1)
    matcher.fit(sample_texts, sample_labels)
    assert matcher.match('добрый день') == ('hello', 1)
    assert matcher.match('добрый вечер') == ('hello', 1/3)
    assert matcher.match('добрый') == ('hello', 1/2)


def test_tfidf_matcher():
    new_texts = ['добрый день', 'доброй ночи', 'добрый вечер', 'доброе утро', 'животное хомяк', 'животное пингвин']
    new_labels = ['hello', 'hello', 'hello', 'hello', 'animal', 'animal']
    matcher = matchers.TFIDFMatcher(threshold=0.3, text_normalization='fast_lemmatize')
    matcher.fit(new_texts, new_labels)
    assert matcher.match('добрый день') == ('hello', 1)
    assert matcher.match('добрый упоротыш') == NO_MATCH
    assert matcher.match('добрый хомяк')[0] == 'animal'
    assert matcher.match('животное собака')[0] == 'animal'


@pytest.mark.parametrize('weights', [None, (1, 1), (0.2, 0.8)])
def test_average_matcher(weights):
    matcher = matchers.WeightedAverageMatcher(
        matchers=[matchers.make_matcher('exact'), matchers.JaccardMatcher()],
        threshold=0.1, weights=weights,
    )
    matcher.fit(sample_texts, sample_labels)
    weights = weights or (0.5, 0.5)
    w0, w1 = weights[0] / sum(weights), weights[1] / sum(weights)
    assert matcher.match('добрый день') == ('hello', 1)
    assert matcher.match('добрый вечер') == ('hello', 0 * w0 + 1/3 * w1)
    assert matcher.match('добрый') == ('hello', 0 * w0 + 1/2 * w1)


def test_max_matcher():
    matcher = matchers.MaxMatcher(
        matchers=[matchers.make_matcher('exact'), matchers.JaccardMatcher()],
        threshold=0.1,
    )
    matcher.fit(sample_texts, sample_labels)
    assert matcher.match('добрый день') == ('hello', 1)
    assert matcher.match('добрый вечер') == ('hello', 1/3)
    assert matcher.match('добрый') == ('hello', 1/2)


class PrefixModel:
    """ Its main goal is to mimic interface of scikit-learn models """
    def __init__(self):
        self.classes_ = []
        self._x = []
        self._y = []

    def fit(self, X, y):
        self.classes_ = sorted(set(y))
        self._y2i = {val: i for i, val in enumerate(self.classes_)}
        self._x = X
        self._y = y

    def longest_common_prefix(self, lhs, rhs):
        if lhs == rhs == '':
            return 0.0
        result = 0
        for left, right in zip(lhs, rhs):
            if left == right:
                result += 1
            else:
                break
        return result * 2.0 / (len(lhs) + len(rhs))

    def predict_proba(self, X):
        scores = np.zeros((len(X), len(self.classes_)), dtype=np.float)
        for i, text in enumerate(X):
            for x, y in zip(self._x, self._y):
                scores[i, self._y2i[y]] = max(scores[i, self._y2i[y]], self.longest_common_prefix(x, text))
        return scores


def test_model_matcher():
    matcher = matchers.ModelBasedMatcher(model=PrefixModel())
    matcher.fit(sample_texts, sample_labels)
    assert matcher.match('добрый день') == ('hello', 1)
    assert matcher.match('добрый д') == ('hello', 8 / ((8 + 11) / 2))
    assert matcher.match('добрый') == ('hello', 6 / ((6 + 11) / 2))


@pytest.mark.parametrize('matcher_class', [matchers.W2VMatcher, matchers.WMDMatcher])
def test_vectorized_matcher(matcher_class):
    w2v = {
        k: np.array(v) for k, v in
        {
            'привет': [1, 0, 0],
            'добрый': [0.5, 0.5, 0],
            'злой': [0.45, 0.55, 0],
            'день': [0.1, 0.2, 0.7],
            'ночь': [0.0, 0.4, 0.7],
            'сколько': [0.0, 0.1, 0.9],
            'времени': [0.0, 0.5, 0.5],
        }.items()
    }
    matcher = matcher_class(w2v=w2v)
    matcher.fit(sample_texts, sample_labels)
    assert matcher.match('времени сколько') == ('get_time', 1)
    assert matcher.match('абракадабра') == NO_MATCH
    label, score = matcher.match('злой ночь')
    assert label == 'hello'
    assert 0.95 < score < 0.99


def test_scores_aggregation():
    matcher = matchers.JaccardMatcher(threshold=0.1)
    matcher.fit(sample_texts, sample_labels)
    assert matcher.aggregate_scores('добрый день') == {'hello': 1}
    assert matcher.aggregate_scores('добрый день', use_threshold=False) == {'hello': 1, 'get_time': 0}
    assert matcher.aggregate_scores('привет сколько времени') == {'hello': 1/3, 'get_time': 2/3}


def test_regex_matcher():
    more_texts = sample_texts + ['.*врем.*']
    more_labels = sample_labels + ['get_time']

    matcher = matchers.RegexMatcher(add_end=False)
    matcher.fit(more_texts, more_labels)
    assert matcher.aggregate_scores('привет мир') == {'hello': 1}
    assert matcher.aggregate_scores('привет расскажи время') == {'get_time': 1, 'hello': 1}

    matcher = matchers.RegexMatcher(add_end=True)
    matcher.fit(more_texts, more_labels)
    assert matcher.aggregate_scores('привет мир') == {}
    assert matcher.aggregate_scores('расскажи время') == {'get_time': 1}
    assert matcher.aggregate_scores('привет расскажи время') == {'get_time': 1}



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tgalice",
    version="0.2.10",
    author="David Dale",
    author_email="dale.david@mail.ru",
    description="Yet another common wrapper for Alice skills and Facebook/Telegram bots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avidale/tgalice",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'attrs',
        'flask',
        'pymessenger',
        'pymorphy2',
        'pyTelegramBotAPI',
        'pyyaml',
        'requests',  # it is needed by other packages, and somehow it is installed wrongly
        'textdistance',
    ],
    extras_require={
        'rumorph': ['pymorphy2[fast]', 'pymorphy2-dicts-ru'],  # todo: move them out of main requirements
        'server': ['flask', 'pymessenger', 'pyTelegramBotAPI'],  # todo: move them out of main requirements
        'w2v':  ['numpy', 'pyemd'],
    }
)



"""
This package implements Yandex.Dialogs response protocol.
The official documentation is available at
https://yandex.ru/dev/dialogs/alice/doc/protocol-docpage/#response
"""

import attr
import copy

from typing import Union, Dict, List, Optional
from tgalice.utils.serialization import Serializeable, list_converter


class CARD_TYPES:
    BIG_IMAGE = 'BigImage'
    ITEMS_LIST = 'ItemsList'


@attr.s
class Button(Serializeable):
    title: str = attr.ib()
    payload = attr.ib()
    url: str = attr.ib()
    hide: bool = attr.ib()


@attr.s
class Card(Serializeable):
    type: str = attr.ib()


@attr.s
class CardButton(Serializeable):
    text: str = attr.ib(default=None)
    url: str = attr.ib(default=None)
    payload = attr.ib(default=None)


@attr.s
class BigImage(Card):
    type: str = attr.ib(default=CARD_TYPES.BIG_IMAGE, init=False)
    image_id: str = attr.ib()
    title: str = attr.ib(default=None)
    description: str = attr.ib(default=None)
    button: CardButton = attr.ib(default=None, converter=CardButton.from_dict)


@attr.s
class ItemsListHeader(Serializeable):
    text: str = attr.ib()


@attr.s
class ItemsListFooter(Serializeable):
    text: str = attr.ib()
    button: Optional[CardButton] = attr.ib(default=None, converter=CardButton.from_dict)


@attr.s
class ItemsListItem(Serializeable):
    image_id: str = attr.ib(default=None)
    title: str = attr.ib(default=None)
    description: str = attr.ib(default=None)
    button: CardButton = attr.ib(default=None, converter=CardButton.from_dict)


@attr.s
class ItemsList(Card):
    type: str = attr.ib(default=CARD_TYPES.ITEMS_LIST, init=False)
    header: Optional[ItemsListHeader] = attr.ib(converter=ItemsListHeader.from_dict, default=None)
    items: List[ItemsListItem] = attr.ib(converter=list_converter(ItemsListItem), factory=list)
    footer: Optional[ItemsListFooter] = attr.ib(converter=ItemsListFooter.from_dict, default=None)


def card_converter(data):
    if not data:
        return None
    if isinstance(data, Card):
        return data
    new_data = copy.deepcopy(data)
    card_type = new_data['type']
    del new_data['type']

    if card_type == CARD_TYPES.BIG_IMAGE:
        return BigImage.from_dict(new_data)
    if card_type == CARD_TYPES.ITEMS_LIST:
        return ItemsList.from_dict(new_data)


@attr.s
class Response(Serializeable):
    text: str = attr.ib()
    tts: str = attr.ib(default=None)
    buttons: List[Button] = attr.ib(converter=list_converter(Button), factory=list)
    card: Optional[Union[BigImage, ItemsList]] = attr.ib(converter=card_converter, default=None)
    end_session: bool = attr.ib(default=False)


@attr.s
class YandexResponse(Serializeable):
    response: Response = attr.ib(converter=Response.from_dict)
    session_state: Optional[Dict] = attr.ib(default=None)
    user_state_update: Optional[Dict] = attr.ib(default=None)
    session = attr.ib(default=None)
    version: str = attr.ib(default='1.0')



from tgalice.nlg import reply_markup, controls


class Response:
    def __init__(self, text,
                 suggests=None, commands=None, voice=None, links=None,
                 image_id=None, image_url=None, sound_url=None,
                 gallery=None, image=None,
                 user_object=None, raw_response=None,
                 confidence=0.5, label=None,
                 rich_text=None,
                 ):
        self.text = text
        self.suggests = suggests or []
        self.commands = commands or []
        self.voice = voice if voice is not None else text
        self.links = links or []
        self.updated_user_object = user_object
        self.confidence = confidence
        self.image_id = image_id
        self.image_url = image_url  # todo: support them in Facebook as well
        self.sound_url = sound_url
        self.gallery = gallery
        assert self.gallery is None or isinstance(self.gallery, controls.Gallery)
        self.image = image
        assert self.image is None or isinstance(self.image, controls.BigImage)
        self.raw_response = raw_response
        self.label = label
        if rich_text:
            self.set_text(rich_text)

    @property
    def user_object(self):
        # make it readonly, for clarity
        return self.updated_user_object

    def set_rich_text(self, rich_text):
        parser = reply_markup.TTSParser()
        try:
            parser.feed(rich_text)
        except ValueError as e:
            raise ValueError('Got error "{}" while parsing text "{}"'.format(e, rich_text))
        parser.close()
        self.text = parser.get_text()
        self.voice = parser.get_voice()
        self.links.extend(parser.get_links())
        return self

    def set_text(self, text_and_voice):
        # this method name is deprecated
        return self.set_rich_text(rich_text=text_and_voice)



import logging
import os

from collections.abc import Mapping

logger = logging.getLogger(__name__)


def get_mongo_or_mock(mongodb_uri=None, mongodb_uri_env_name='MONGODB_URI'):
    if mongodb_uri is None:
        mongodb_uri = os.environ.get(mongodb_uri_env_name)
    if mongodb_uri:
        from pymongo import MongoClient
        mongo_client = MongoClient(mongodb_uri)
        mongo_db = mongo_client.get_default_database()
    else:
        logging.warning('Did not found mongodb uri, trying to load mongomock instead. '
                        'This storage is not persistable and may cause problems in production; '
                        'please use it for testing only.')
        import mongomock
        mongo_client = mongomock.MongoClient()
        mongo_db = mongo_client.db
    return mongo_db


def get_boto_s3(
        service_name='s3',
        endpoint_url='https://storage.yandexcloud.net',
        aws_access_key_id=None,
        aws_access_key_id_env_name='AWS_ACCESS_KEY_ID',
        aws_secret_access_key=None,
        aws_secret_access_key_env_name='AWS_SECRET_ACCESS_KEY',
        region_name='ru-central1',
):
    import boto3
    session = boto3.session.Session()
    s3 = session.client(
        service_name=service_name,
        endpoint_url=endpoint_url,
        aws_access_key_id=aws_access_key_id or os.environ[aws_access_key_id_env_name],
        aws_secret_access_key=aws_secret_access_key or os.environ[aws_secret_access_key_env_name],
        region_name=region_name,
    )
    return s3


def replace_dotted_keys(data, dot_symbol='~'):
    """ Replace dots in dict keys with other symbols, to comply with Pymongo checks """
    if isinstance(data, Mapping):
        result = {}
        for key, value in data.items():
            new_key = key
            if isinstance(key, str):
                new_key = new_key.replace('.', dot_symbol)
            result[new_key] = replace_dotted_keys(value, dot_symbol=dot_symbol)
        return result
    elif isinstance(data, list):
        return [replace_dotted_keys(item, dot_symbol=dot_symbol) for item in data]
    else:
        return data
